'''
The command line interface (CLI).
'''
import sys
import argparse
import asyncio
from paipe import util
from paipe.models import PaipeContext
from .profiles import list_profiles
from paipe.main import run_agent

util.patch_video_mimetype()

GLOBAL_ACTION_ARGS = [
    '-h', '--help',
    '-v', '--verbose',
]

GLOBAL_ARGS = GLOBAL_ACTION_ARGS + [
    '--version'
]

SUB_COMMANDS = [
    'call',
    'op'
]

def build_call_parser(parser):
    parser.add_argument('-S', '--system-prompt', 
                       type=str,
                       help='Specify the system prompt')
    parser.add_argument('--stream', 
                       dest='stream',
                       action='store_true',
                       help='Enable stream mode (default)')
    parser.add_argument('--no-stream',
                       dest='stream', 
                       action='store_false',
                       help='Disable stream mode')
    parser.set_defaults(stream=True)
    parser.add_argument('-P', '--profile',
                        type=str,
                        help='Specify the profile to use')
    parser.add_argument('--json',
                        type=str,
                        default=None,
                        help='The JSON Schema for the result, implies --no-stream')
    parser.add_argument('--file',
                       type=str,
                       help='Read text from a file and append to the prompt')
    parser.add_argument('--model',
                       type=str,
                       default=None,
                       help='Specify the model to use(Overrides the profile)')
    parser.add_argument('-A', '--attach',
                       type=str,
                       action='append',
                       help='Add a file as attachment')
    parser.add_argument('--list',
                        nargs='?',
                        type=str,
                        default=None,
                        const=True,
                        help='List available profiles with optional prefix specified')
    parser.add_argument('-e', '--extract-code-block',
                        nargs='?',
                        type=str,
                        default=None,
                        const=True,
                        help='Extract last code block matched from the resposne, with optional language specified, implies --no-stream')
    parser.add_argument('-o', '--operation',
                        type=str,
                        action=util.DeprecatedAction,
                        deprecated='Deprecated, use `paipe op` instead',
                        help='Perform a specific operation')
    parser.add_argument('--usage',
                        action='store_true',
                        help='Show usage information in stderr.')
    parser.add_argument('prompt',
                       nargs='*',
                       type=str,
                       default='',
                       help='The prompt to process')


def build_parser(with_sub_parser: bool, default_call_parser: bool) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog='paipe',
                                     description='A CLI tool for accessing the LLM API in the terminal.')

    if with_sub_parser:
        sub_parsers = parser.add_subparsers(dest='command',
                                            title=' Commands',
                                            metavar='',
                                            help='')
        call_parser = sub_parsers.add_parser('call', help='Default command(`call` can be skiped), directly call an LLM API.')
        build_call_parser(call_parser)
        command_parser = sub_parsers.add_parser('op', help='Perform a specific operation')
        from .operations.subcli import build_command_parser
        build_command_parser(command_parser)

    # Bind sub parser 'call' as root parser 
    if default_call_parser:
        build_call_parser(parser)

    parser.add_argument('-v', '--verbose',
                       action='store_true',
                       help='Enable verbose output')
    parser.add_argument('--version',
                       action='store_true',
                       help='Show version information and quit')
    return parser


def parse_args(cmd_args: list) -> argparse.Namespace:
    # Only deal with global args
    if (set(cmd_args) & set(GLOBAL_ACTION_ARGS) != set()):
        parser = build_parser(True, True)
        args = parser.parse_args(cmd_args)
    else:
        # if --help is here, it will be parsed by global parser
        parser = build_parser(False, True)
        args, __ = parser.parse_known_args(cmd_args)
        # Check if the first argument is a sub command
        if (args.prompt and args.prompt[0] in SUB_COMMANDS):
            parser = build_parser(True, False)
            args = parser.parse_args(cmd_args)
    return args


def handle_global_args(args: argparse.Namespace):
    if args.verbose:
        util.set_verbose(True)
    else:
        util.set_verbose(False)
    if args.version:
        from . import __version__
        print(f'paipe {__version__}')
        sys.exit(0)


def handle_call(args):
    if args.list:
        profiles = list_profiles(args.list)
        sys.exit(0)
    context_dict = {
        'stream': args.stream,
        'prompt': ' '.join(args.prompt),
        'json_schema': args.json,
        'extract_code_block': args.extract_code_block,
        'model': args.model,
        'attachments': [],
        'usage': args.usage,
    }
    if args.file:
        with open(args.file, 'r') as f:
            context_dict['input_text'] = f.read()
    elif not sys.stdin.isatty():
            context_dict['input_text'] = sys.stdin.read()
    else:
        if not context_dict['prompt']:
            print('No prompt or input via stdin or a file with --file provided')
            sys.exit(1)        

    if args.profile:
        context_dict['profile'] = args.profile

    if args.system_prompt:
        context_dict['system_prompt'] = args.system_prompt
    else:
        context_dict['system_prompt'] = None

    if args.attach:
        context_dict['attachments'].extend(util.to_attachment_pairs(args.attach))

    if args.operation:
        from .operations import handle_operation
        handle_operation(args.operation, context_dict)
    else:
        context = PaipeContext.model_validate(context_dict)
        asyncio.run(run_agent(context))


def main():
    args = parse_args(sys.argv[1:])
    handle_global_args(args)
    if getattr(args, 'command', None) is None or args.command == 'call':
        handle_call(args)
    elif args.command == 'op':
        from .operations.base import handle_operation
        handle_operation(args)
    else:
        print(f'Unknown command: {args.command}')
        sys.exit(1)


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        util.logger.exception(e)
        if hasattr(e, 'body'):
            util.logger.error(e.body)


import sys
import argparse

def parse_args(args):
    parser = argparse.ArgumentParser(description='A test program.')
    parser.add_argument('--arg1', type=str, help='Argument 1')
    parser.add_argument('--arg2', type=int, help='Argument 2')
    subs = parser.add_subparsers(dest='command')
    sub1 = subs.add_parser('sub1', help='Subcommand 1')
    sub1.add_argument('--subarg1', type=str, help='Subargument 1')
    sub1.add_argument('prompt',
                       nargs='*',
                       type=str,
                       default='',
                       help='The prompt to process')

    return parser.parse_args(args)

args = parse_args(sys.argv[1:])
print(args)


from . import archive


def handle_operation(operation, context_dict):
    if operation == 'filelist-to-md':   
        print(archive.archive(context_dict['input_text']))
    else:
        print(f'Unknown operation: {operation}')



'''
Convert a list of files into a markdown of file contents.
'''
import os
from typing import List, Tuple, Literal, TypeVar

WrapType = Literal['html', 'markdown', 'none']


def split_to_filelist(output):
    filelist = []
    for i in output.split():
        file = i.strip()
        if not file:
            continue
        filelist.append(file)
    return filelist


def wrap_html_comment(filename, content):
    '''
    Wrap content in html comment, aka. <!--- comment --->
    '''
    return f'<!-- begin {filename} -->\n{content}\n<!-- end {filename} -->\n\n'


def wrap_markdown_comment(filename, content):
    '''
    Wrap content in markdown comment, aka. [comment]: #
    '''
    return f'[begin {filename}]: #\n{content}\n[end {filename}]: #\n\n'


def wrap_none(__, content):
    return content


def get_wrap_method(wrap: WrapType):
    if wrap == 'markdown':
        return wrap_markdown_comment
    elif wrap == 'none':
        return wrap_none
    else:
        return wrap_html_comment


def archive_to_markdown(files,
                        content_list: List[Tuple[str, str]],
                        wrap_method=None
    ):
    wrap_method = wrap_method or wrap_none
    markdown = ''
    for filename in files:
        if os.path.isfile(filename):
            with open(filename, 'r') as fd:
                markdown += wrap_method(filename, fd.read())
    if content_list:
        for (filename, content) in content_list:
            markdown += wrap_method(filename, content)
    return markdown


def archive(output,
            filelist: list | None=None,
            use_stdin_as: str='list',
            wrap: Literal['html', 'markdown', 'none']=None
    ):
    '''
    Convert the filelist to content markdown.
    '''
    if filelist is None:
        filelist = []
    content_list = []
    if use_stdin_as == 'list':
        pipe_filelist = split_to_filelist(output)
        filelist.extend(pipe_filelist)
    else:
        content_list.append(('STDIN', output))
    wrap_method = get_wrap_method(wrap)
    return archive_to_markdown(filelist, content_list, wrap_method=wrap_method)


from pydantic import BaseModel, Field


class PaipeContext(BaseModel):
    profile: str = Field(default='default', description='The profile name')
    system_prompt: str | None = Field(default=None, description='The system prompt')
    stream: bool = Field(default=True, description='Enable stream mode')
    input_text: str = Field(default='', description='The input text')
    prompt: str = Field(default='', description='The prompt')
    json_schema: str | None = Field(default=None, description='The JSON schema for the result')
    extract_code_block: bool | str | None = Field(default=None, description='Extract code block')
    model: str | None = Field(default=None, description='The model name')
    attachments: list = Field(default=[], description='The attachments')
    usage: bool = Field(default=False, description='Show usage')


from typing import Any, Literal
import pydantic_ai.models.openai
from pydantic_ai.models.openai import (
    ModelMessage,
    ModelSettings,
    ModelRequestParameters,
    OpenAIModelSettings,
    chat,
    AsyncStream,
    ChatCompletionChunk,
    NOT_GIVEN,
    APIStatusError,
    ModelHTTPError
)


def non_annotated(model_settings: ModelSettings) -> dict[str, Any]:
    """Extract non annotated key-value pairs from model_settings."""
    return {k: v for (k, v) in model_settings.items() if k not in OpenAIModelSettings.__annotations__}


class OpenAIModel(pydantic_ai.models.openai.OpenAIModel):
    '''
    Subclass of OpenAIModel to allow for custom model settings.
    '''
    async def _completions_create(
        self,
        messages: list[ModelMessage],
        stream: bool,
        model_settings: OpenAIModelSettings,
        model_request_parameters: ModelRequestParameters,
    ) -> chat.ChatCompletion | AsyncStream[ChatCompletionChunk]:
        tools = self._get_tools(model_request_parameters)

        # standalone function to make it easier to override
        if not tools:
            tool_choice: Literal['none', 'required', 'auto'] | None = None
        elif not model_request_parameters.allow_text_result:
            tool_choice = 'required'
        else:
            tool_choice = 'auto'

        if hasattr(self, '_map_messages'):
            openai_messages = await self._map_messages(messages)
        else:
            openai_messages: list[chat.ChatCompletionMessageParam] = []
            for m in messages:
                async for msg in self._map_message(m):
                    openai_messages.append(msg)
        try:
            return await self.client.chat.completions.create(
                model=self._model_name,
                messages=openai_messages,
                n=1,
                parallel_tool_calls=model_settings.get('parallel_tool_calls', NOT_GIVEN),
                tools=tools or NOT_GIVEN,
                tool_choice=tool_choice or NOT_GIVEN,
                stream=stream,
                stream_options={'include_usage': True} if stream else NOT_GIVEN,
                max_tokens=model_settings.get('max_tokens', NOT_GIVEN),
                temperature=model_settings.get('temperature', NOT_GIVEN),
                top_p=model_settings.get('top_p', NOT_GIVEN),
                timeout=model_settings.get('timeout', NOT_GIVEN),
                seed=model_settings.get('seed', NOT_GIVEN),
                presence_penalty=model_settings.get('presence_penalty', NOT_GIVEN),
                frequency_penalty=model_settings.get('frequency_penalty', NOT_GIVEN),
                logit_bias=model_settings.get('logit_bias', NOT_GIVEN),
                reasoning_effort=model_settings.get('openai_reasoning_effort', NOT_GIVEN),
                **non_annotated(model_settings),
            )
        except APIStatusError as e:
            if (status_code := e.status_code) >= 400:
                raise ModelHTTPError(status_code=status_code, model_name=self.model_name, body=e.body) from e
            raise


import sys
import json
from pathlib import Path
import pydantic
import dydantic
import pydantic_ai.models
from pydantic_ai import Agent
from pydantic_ai.messages import BinaryContent
from .models import PaipeContext
from .util import (
    logger,
    import_module,
    extract_markdown_code_blocks,
    show_json_usage
)
from . profiles import get_profile


def import_model_module(name: str):
    module = import_module('paipe.providers', name)
    if not module:
        module = import_module('pydantic_ai.models', name)
    return module


def import_provider_module(name: str):
    return import_module('pydantic_ai.providers', name)


def get_agent_model_cls(module):
    for name in dir(module):
        obj = getattr(module, name)
        if hasattr(obj, '__mro__') and pydantic_ai.models.Model in obj.__mro__[1:]:
            return obj
    return None


def get_agent_provider_cls(module):
    for name in dir(module):
        obj = getattr(module, name)
        if hasattr(obj, '__mro__') and pydantic_ai.providers.Provider in obj.__mro__[1:]:
            return obj
    return None


def get_agent_model(model_name: str, protocol: str, provider: str, **profile):
    model_params = {}
    model_cls = get_agent_model_cls(import_model_module(protocol))
    provider_module = import_provider_module(provider)
    if provider_module:
        provider_cls = get_agent_provider_cls(provider_module)
        if provider_cls:
            provider_params = {}
            for key in provider_cls.__init__.__annotations__:
                if key == 'return':
                    continue
                if key in profile:
                    provider_params[key] = profile.pop(key)
            provider_instance = provider_cls(**provider_params)
            model_params['provider'] = provider_instance
    model_params.update(profile)
    return model_cls(model_name, **model_params)


def process_prompt(full_prompt: str, attachments: list | None = None) -> str | list:
    if not attachments:
        result = full_prompt
    else:
        result = [
            full_prompt
        ]
        for (media_type, data) in attachments:
            content = BinaryContent(
                data=data,
                media_type=media_type
            )
            result.append(content)
        else:
            pass
    return result


async def run_agent(context: PaipeContext):
    profile = get_profile(context.profile)
    if profile is None:
        print(f"Profile {context.profile} is not available in the profile.")
        sys.exit(1)

    protocol = profile.pop('protocol', None) or profile.pop('provider', None) or 'openai'
    provider = profile.pop('provider', None) or 'openai'
    model = profile.pop('model', None) or ''
    profile_system_prompt = profile.pop('system_prompt', None)
    model_settings =  profile.pop('model_settings', None)

    agent_params = {
        "system_prompt": context.system_prompt or profile_system_prompt or (),
        "model_settings": model_settings
    }
    if context.json_schema:
        agent_params['result_type'] = dydantic.create_model_from_schema(
                json.loads(context.json_schema))
        context.stream = False
    elif context.extract_code_block:
        context.stream = False
    logger.debug(f'[model to use] {context.model or model}')

    agent = Agent(get_agent_model(context.model or model, protocol, provider, **profile),
                  **agent_params)
    full_prompt  = ''
    if context.prompt:
        full_prompt += f'{context.prompt}\n'
    if context.input_text:
        full_prompt += f'{context.input_text}\n'

    processed_prompt = process_prompt(full_prompt,
                                      attachments=context.attachments)

    if context.stream:
        async with agent.run_stream(processed_prompt) as response:
            async for delta in response.stream_text(delta=True):
                print(delta, end='', flush=True)
        print()
        if context.usage:
            show_json_usage(response.usage())
    else:
        result = await agent.run(processed_prompt)
        if context.json_schema and isinstance(result.data, pydantic.BaseModel):
            print(result.data.model_dump_json())
        elif context.extract_code_block and isinstance(result.data, str):
            language = '' if context.extract_code_block is True else context.extract_code_block
            code_blocks = extract_markdown_code_blocks(
                result.data,
                language=language
            )
            if not code_blocks:
                logger.warning(f"No {language} code block detected.")
            print(code_blocks[-1] if code_blocks else '')
        else:
            print(result.data)
        if context.usage:
            show_json_usage(result.usage())



import sys

def get_stdio():
    if not sys.stdin.isatty():
        return sys.stdin.read()


def handle_operation(args):
    if args.operation == 'archive':
        from . archive import archive
        print(
            archive(get_stdio() or '', args.filelist, use_stdin_as=args.stdin)
        )
 


import sys
import json
import base64
import mimetypes
import logging
import importlib
import re
import argparse
from typing import Generator
import pydantic_ai.result

logger = logging.getLogger('paipe')
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
logger.addHandler(handler)


def set_verbose(verbose: bool):
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


class DeprecatedAction(argparse.Action):
    def __init__(self, option_strings, deprecated=False, help=None, **kwargs):
        if help is not None:
            if deprecated:
                help += ' (DEPRECATED)'
        super().__init__(option_strings, help=help, **kwargs)
        self.deprecated = deprecated

    def __call__(self, parser, namespace, values, option_string=None):
        if self.deprecated:
            logger.warning(
                self.deprecated if isinstance(self.deprecated, str)\
                      else f'{self.option_strings[0]} is deprecated.'
                )
        setattr(namespace, self.dest, values)


def file_as_data_url(file_path: str) -> str:
    with open(file_path, 'rb') as f:
        data = f.read()
    mime_type, _ = mimetypes.guess_type(file_path)
    if mime_type is None:
        mime_type = 'application/octet-stream'
    data_type = 'text'
    if mime_type.startswith('image/'):
        data_type = 'image'
    elif mime_type.startswith('video/'):
        data_type = 'video'
    return (data_type, f'data:{mime_type};base64,{base64.b64encode(data).decode("utf-8")}')


def to_attachment_pairs(file_path_list: list[str]) -> Generator[tuple[bytes, str], None, None]:
    for file_path in file_path_list:
        data = open(file_path, 'rb').read()
        mime_type, _ = mimetypes.guess_type(file_path)
        if mime_type is None:
            mime_type = 'application/octet-stream'
        yield (mime_type, data)


def import_module(parent_module: str, name: str):
    '''
    Import a module from a parent module.
    '''
    try:
        return importlib.import_module(f'{parent_module}.{name}')
    except ImportError as e:
        return None


def find_cls(module, base_class):
    '''
    Find a class in a module that inherits from a base class.
    '''
    for name in dir(module):
        obj = getattr(module, name)
        if hasattr(obj, '__mro__') and base_class in obj.__mro__[1:]:
            return obj
    return None


def extract_markdown_code_blocks(markdown: str, language: str = '') -> list:
    """
    Extract code blocks from a markdown document.

    Args:
        markdown: The markdown text to process
        language: Optional language filter (case insensitive)

    Returns:
        A list of extracted code blocks
    """
    pattern = r'```([\w+-]*)\n(.*?)\n```'
    matches = re.findall(pattern, markdown, re.DOTALL)
    if language:
        return [code for lang, code in matches
                if lang.lower() == language.lower()]
    else:
        return [code for _, code in matches]


def patch_video_mimetype():
    @property
    def is_image(self) -> bool:
        """Return `True` if the media type is an image type."""
        return self.media_type.startswith(('image/', 'video/'))
    from pydantic_ai.messages import BinaryContent
    BinaryContent.is_image = is_image
    logger.debug("Patched pydantic_ai.messages.BinaryContent.is_image")


def show_json_usage(usage: pydantic_ai.result.Usage,
                    file=None):
    if file is None:
        file = sys.stderr
    print("Usage:", json.dumps(usage.__dict__), file=file)